
import aiinterface.AIInterface;
import java.util.Vector;
import aiinterface.CommandCenter;
import enumerate.Action;
import simulator.Simulator;
import struct.CharacterData;
import struct.FrameData;
import struct.GameData;
import struct.Key;
import struct.MotionData;

public class UtalFighter implements AIInterface {

	Key m_inputKey;
	boolean m_player;
	
	private FrameData m_FrameData;
	
	CommandCenter m_commandCenter;
	Simulator m_simulator;
	
	private Node rootNode;
	
	CharacterData m_characterData;
	CharacterData enemyCharacterData;
	
	private Vector<MotionData> _m_motionData;
	private Vector<MotionData> enemyMotionData;
	
	MyStateMachine MFSM;
	
	Information mySelf;
	Information enemy;
	
	GameData gd;
	
	@Override
	public String getCharacter(){
		return CHARACTER_ZEN;
	}
	
	@Override
	public void close() {
		// TODO Auto-generated method stub

	}

	@Override
	public void getInformation(FrameData framedata) {
		
		this.m_FrameData = framedata;
		this.m_commandCenter.setFrameData(m_FrameData, m_player);
		
		if(this.m_FrameData != null) {
			this.m_characterData = this.m_FrameData.getCharacter(this.m_player);
			this.enemyCharacterData = this.m_FrameData.getCharacter(!this.m_player);
			
			if(m_characterData != null) {
				if(mySelf.maxHP == 0) {
					this.mySelf.maxHP = this.m_characterData.getHp();
				}
				this.mySelf.currentHP = this.m_characterData.getHp();
				this.mySelf.distanceBetween = this.m_FrameData.getDistanceX();
				this.mySelf.currentEnergy = this.m_characterData.getEnergy();
				this.mySelf.posX = this.m_characterData.getLeft();
			}	
			
			if(enemyCharacterData != null) {
				if(enemy.maxHP == 0) {
					this.enemy.maxHP = this.enemyCharacterData.getHp();
				}
				this.enemy.currentHP = this.enemyCharacterData.getHp();
				this.enemy.distanceBetween = 0; //Ya tengo la distancia en mi informacion
				this.enemy.currentEnergy = this.m_characterData.getEnergy();
				this.enemy.posX = 0;
			}
		}
		

	}

	@Override
	public int initialize(GameData gameData, boolean player) {
		this.m_inputKey = new Key();
		this.m_FrameData = new FrameData();
		this.m_player = player;
		this.m_commandCenter = new CommandCenter();
		this.m_simulator = gameData.getSimulator();
		this.gd = gameData;
		
		this.mySelf = new Information();
		this.enemy = new Information();
		
		if(this.m_simulator == null) {
			System.out.println("El simulador es nulo");
			
		}
		
		//init State machine
		MFSM = new MyStateMachine(gd, this.m_player);
		
		MFSM.initialState = MFSM.states.get(0);
		MFSM.activeState = MFSM.initialState;
		
		return 0;
	}

	@Override
	public Key input() {
		return this.m_inputKey;
	}

	@Override
	public void processing() {
		
		if(canProcess()) {
			
			//System.out.println("Distance = " + distance);
			
			if(this.m_commandCenter.getSkillFlag()) {
				m_inputKey = this.m_commandCenter.getSkillKey();
				
			}else {
				m_inputKey.empty();
				this.m_commandCenter.skillCancel();
				
				//this.m_commandCenter.commandCall(MFSM.update( this.mySelf , this.enemy).toString());
				//Action action = MonteCarlo_ts.subProcesing(m_simulator, m_FrameData, m_player, gd, m_commandCenter);
				Action action = MFSM.update(this.m_characterData, this.enemyCharacterData, this.m_simulator, this.m_FrameData, this.m_player, this.gd, this.m_commandCenter);
				this.m_commandCenter.commandCall(action.toString());
				
				
			}
		}
	}
	
	public boolean canProcess() {
		return !m_FrameData.getEmptyFlag() && m_FrameData.getRemainingTimeMilliseconds() > 0;
	}

	@Override
	public void roundEnd(int arg0, int arg1, int arg2) {
		// TODO Auto-generated method stub

	}
	
	/*
	 * actionAir =
		        new Action[] {Action.AIR_GUARD, Action.AIR_A, Action.AIR_B, Action.AIR_DA, Action.AIR_DB,
		            Action.AIR_FA, Action.AIR_FB, Action.AIR_UA, Action.AIR_UB, Action.AIR_D_DF_FA,
		            Action.AIR_D_DF_FB, Action.AIR_F_D_DFA, Action.AIR_F_D_DFB, Action.AIR_D_DB_BA,
		            Action.AIR_D_DB_BB};
		
		actionGround =
		        new Action[] {Action.STAND_D_DB_BA, Action.BACK_STEP, Action.FORWARD_WALK, Action.DASH,
		            Action.JUMP, Action.FOR_JUMP, Action.BACK_JUMP, Action.STAND_GUARD,
		            Action.CROUCH_GUARD, Action.THROW_A, Action.THROW_B, Action.STAND_A, Action.STAND_B,
		            Action.CROUCH_A, Action.CROUCH_B, Action.STAND_FA, Action.STAND_FB, Action.CROUCH_FA,
		            Action.CROUCH_FB, Action.STAND_D_DF_FA, Action.STAND_D_DF_FB, Action.STAND_F_D_DFA,
		            Action.STAND_F_D_DFB, Action.STAND_D_DB_BB};
		
	    spSkill = Action.STAND_D_DF_FC;
	 */
	

}
